/*******************************/
/* mbone-reflector-v4m-v6m by  */
/*    Konstantin Kabassanov    */
/*         LIP6, Paris         */
/*  konstantin@kabassanov.com  */
/*******************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/signal.h>
#include <sys/param.h>
#include <errno.h>
#include <netdb.h>
#include <arpa/inet.h>

#define MSGBUF_SIZE 32768

/* IPv4 RTP Statistics */
int mcdatatotalbytesrecv4 = 0;
int mcdatatotalbytessent4 = 0;
int mcdatarecvfromcalls4 = 0;
int mcdatasendtocalls4 = 0;

/* IPv6 RTP Statistics */
int mcdatatotalbytesrecv = 0;
int mcdatatotalbytessent = 0;
int mcdatarecvfromcalls = 0;
int mcdatasendtocalls = 0;

/* IPv4 RTCP Statistis */
int mcrtcptotalbytesrecv4 = 0;
int mcrtcptotalbytessent4 = 0;
int mcrtcprecvfromcalls4 = 0;
int mcrtcpsendtocalls4 = 0;

/* IPv6 RTCP Statistics */
int mcrtcptotalbytesrecv = 0;
int mcrtcptotalbytessent = 0;
int mcrtcprecvfromcalls = 0;
int mcrtcpsendtocalls = 0;

int programshutdown (void);

int
main (argc, argv)
     int argc;
     char **argv;
{
  /*define variables*/
  
  int mcdatarecvfd4;   //multicast data receive socket IPv4 
  int mcdatarecvfd;    //multicast data receive socket IPv6
  
  int mcrtcprecvfd4;   //multicast rtcp receive socket IPv4
  int mcrtcprecvfd;    //multicast rtcp receive socket IPv6
  
  int sendfd4;   //local send socket IPv4
  int sendfd;    //local send socket IPv6
  
  u_char ttl4;   //time to live IPv4
  int ttl;       //time to live IPv6
  
  struct sockaddr_in mcdataaddr4; //reception IPv4 sockaddr  
  struct sockaddr_in6 mcdataaddr; //reception IPv6 sockaddr
  
  struct sockaddr_in mcrtcpaddr4; //reception IPv4 RTCP sockaddr
  struct sockaddr_in6 mcrtcpaddr; //reception IPv6 RTCP sockaddr
  
  struct sockaddr_in localsendaddr4; //IPv4 address of outgoing packets
  struct sockaddr_in6 localsendaddr; //IPv6 address of outgoing packets
  
  struct sockaddr_in sourceaddr4; //IPv4 address of incoming packets
  struct sockaddr_in6 sourceaddr; //IPv6 address of incoming packets
  
  struct ip_mreq mcdatareq4; //IPv4 join group structure
  struct ipv6_mreq mcdatareq; //IPv6 join group structure
  
  int mcdataaddrlen4; //IPv4 multicast address length
  int mcdataaddrlen;  //IPv6 multicast address length
  
  int mcrtcpaddrlen4; //IPv4 RTCP multicast address length
  int mcrtcpaddrlen;  //IPv6 RTCP multicast address length
  
  int sourceaddrlen4; //IPv4 source address length
  int sourceaddrlen;  //IPv6 source address length
  
  char mcdatarecvbuf4[MSGBUF_SIZE]; //IPv4 multicast receive buffer
  char mcdatarecvbuf[MSGBUF_SIZE];  //IPv6 multicast receive buffer
  
  char mcrtcprecvbuf4[MSGBUF_SIZE]; //IPv4 multicast RTCP receive buffer
  char mcrtcprecvbuf[MSGBUF_SIZE]; //IPv6 multicast RTCP receive buffer
  
  int mcdataport4; //IPv4 multicast port 
  int mcdataport;  //IPv6 multicast port
  
  int mcrtcpport4; //IPv4 multicast RTCP port
  int mcrtcpport;  //IPv6 multicast RTCP port
  
  struct in_addr multicastaddress4; //IPv4 multicast address from command line 
  struct in6_addr multicastaddress; //IPv6 multicast address from command line
  
 
  fd_set readfds; // file descriptors set 
  int maxfds;     // max number of file descriptors
  int nfds;      
  int nr;  // number of bytes read with recvfrom
  int ns;  // number of bytes sent with sendto
  int chksrc;  // asserts if source address is not the reflector's one
  char myhostname[MAXHOSTNAMELEN]; //hostname
  char addressstring4[20];         // IPv4 address string
  char addressstring[80];          // IPv6 address string
  char myhostnameipaddress4[20];   // IPv4 reflector's address string
  char myhostnameipaddress[80];    // IPv6 reflector's address string
  struct hostent  *host, *host4;   // IPv6 and IPv4 hostentries
  int debugon = 0;                 // Debug mode 
  int n2;
  char inputbuf[32];
   
  /*check command line arguments*/
  if (argc < 6) {
    fprintf (stderr, "Usages: mbone-reflector-v4m-v6m [multicast v4 address] [multicast v4  port] [multicast v6 address] [multicast v6 port] [ttl v4] [ttl v6]\n");
    exit(1);
  }
  
  /* IPv4 multicast address */
  if((inet_pton(AF_INET,argv[1],&multicastaddress4.s_addr))!=1){
    printf("bad multicast IPv4 address format\n");
    exit (1);
  }
  
  /* IPv4 multicast ports */
  if (atoi(argv[2]) % 2 == 0) {
    mcdataport4 = atoi(argv[2]);
    mcrtcpport4 = atoi(argv[2]) + 1;
  } else {
    mcdataport4 = atoi(argv[2]) - 1;
    mcrtcpport4 = atoi(argv[2]);
  }
  
  /* IPv6 multicast address */
  if((inet_pton(AF_INET6,argv[3],&multicastaddress))!=1){
    printf("bad multicast address format\n");
    exit (1);
  }
  
  /* IPv6 multicast ports */
  if (atoi(argv[4]) % 2 == 0) {
    mcdataport = atoi(argv[4]);
    mcrtcpport = atoi(argv[4]) + 1;
  } else {
    mcdataport = atoi(argv[4]) - 1;
    mcrtcpport = atoi(argv[4]);
  }

  /* Time to live for IPv4 multicast */
  ttl4 = atoi(argv[5]);
  
  /* Time to live for IPv6 multicast */
  ttl = atoi(argv[6]);

  /* Debug operations */
  if (argv[7] != NULL) {
    if (!strcmp(argv[7],"d")) { 
      printf("command line operation is on\n");
      debugon = 1;
    }
    if (!strcmp(argv[7],"d2")) { 
      printf("debug level 2 is on\n");
      debugon = 2;
    }
    if (!strcmp(argv[7],"d3")) { 
      printf("debug level 3 is on\n");
      debugon = 3;
    }
    if (!strcmp(argv[7],"d4")) { 
      printf("debug level 4 is on\n");
      debugon = 4;
    }
  }
  
  printf("multicastaddress v4 =%s   ",argv[1]);
  printf("mcdataport v4 =%d  ",mcdataport4);
  printf("mcrtcpport v4 =%d\n",mcrtcpport4);
  
  printf("multicastaddress=%s   ",argv[3]);
  printf("mcdataport=%d  ",mcdataport);
  printf("mcrtcpport=%d\n",mcrtcpport);
  
  
  /* Structures initialization */

  /*enter the address/port data into the mcdataaddr structure for IPv4*/
  bzero((char *) &mcdataaddr4, sizeof(mcdataaddr4));
  mcdataaddr4.sin_family=AF_INET;
  mcdataaddr4.sin_addr.s_addr = multicastaddress4.s_addr;
  mcdataaddr4.sin_port = htons(mcdataport4);
  
   
  /*enter the address/port data into the mcdataaddr structure for IPv6*/
  bzero((char *) &mcdataaddr, sizeof(mcdataaddr));
  mcdataaddr.sin6_family=AF_INET6;
  mcdataaddr.sin6_addr = multicastaddress;
  mcdataaddr.sin6_port = htons(mcdataport);
  
  /*enter the address/port data into the mcrtcpaddr structure for IPv4*/
  bzero((char *) &mcrtcpaddr4, sizeof(mcrtcpaddr4));
  mcrtcpaddr4.sin_family=AF_INET;
  mcrtcpaddr4.sin_addr.s_addr = multicastaddress4.s_addr;
  mcrtcpaddr4.sin_port = htons(mcrtcpport4);
  

  /*enter the address/port data into the mcrtcpaddr structure for IPv6*/
  bzero((char *) &mcrtcpaddr, sizeof(mcrtcpaddr));
  mcrtcpaddr.sin6_family=AF_INET6;
  mcrtcpaddr.sin6_addr = multicastaddress;
  mcrtcpaddr.sin6_port = htons(mcrtcpport);
  
  /*figure out the hostname*/
  if (gethostname(myhostname, MAXHOSTNAMELEN) < 0)
    perror("gethostname error");
  printf("myhostname=%s\n", myhostname);
  
  /* Get local IPv4 address */
  if((host4 = gethostbyname2(myhostname,AF_INET)) == NULL) 
    perror("gethostbyname IPv4 failure!");
  
  inet_ntop(AF_INET,(struct in_addr*)host4->h_addr_list[0],myhostnameipaddress4,sizeof(myhostnameipaddress4));
  printf("myhostipaddress=%s\n\n", myhostnameipaddress4);
  
  /* Get local IPv6 address */
  if((host = gethostbyname2(myhostname,AF_INET6)) == NULL) 
    perror("gethostbyname v6 failure!");
  
  inet_ntop(AF_INET6,(struct in6_addr*)host->h_addr_list[0],myhostnameipaddress,sizeof(myhostnameipaddress));
  printf("myhostipaddress=%s\n\n", myhostnameipaddress);
  
  /*enter the address/port data into the localsendaddr structure for IPv4*/
  bzero((char *) &localsendaddr4, sizeof(localsendaddr4));
  localsendaddr4.sin_family=AF_INET;
  inet_pton(AF_INET, myhostnameipaddress4 ,&localsendaddr4.sin_addr.s_addr);
  localsendaddr4.sin_port = htons(0);
  
  /*enter the address/port data into the localsendaddr structure for IPv6*/
  bzero((char *) &localsendaddr, sizeof(localsendaddr));
  localsendaddr.sin6_family=AF_INET6;
  inet_pton(AF_INET6, myhostnameipaddress ,&localsendaddr.sin6_addr);
  localsendaddr.sin6_port = htons(0);
  
  /*enter the address/port data into the mcdatareq structure for IPv4*/
  bzero((char *) &mcdatareq4, sizeof(mcdatareq4));
  mcdatareq4.imr_multiaddr.s_addr=multicastaddress4.s_addr;
  mcdatareq4.imr_interface.s_addr=htonl(INADDR_ANY);


  /*enter the address/port data into the mcdatareq structure for IPv6*/
  bzero((char *) &mcdatareq, sizeof(mcdatareq));
  mcdatareq.ipv6mr_multiaddr=multicastaddress;
  mcdatareq.ipv6mr_interface=htonl(0);
  
  /*get the sendfd socket for IPv4*/
  if ((sendfd4 = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("can't open sendfd socket!");
    exit (1);
  }

  if (bind(sendfd4, (struct sockaddr *) &localsendaddr4, \
	   sizeof(localsendaddr4)) < 0) {
    perror("can't bind localsendaddr v4 to socket!");
    exit(1);
  }
  
  /*get the sendfd socket for IPv6*/
  if ((sendfd = socket(AF_INET6, SOCK_DGRAM, 0)) < 0) {
    perror("can't open sendfd socket!");
    exit (1);
  }
  if (bind(sendfd, (struct sockaddr *) &localsendaddr, \
	   sizeof(localsendaddr)) < 0) {
    perror("can't bind localsendaddr to socket!");
    exit(1);
  }
  
  /*get a mcdatarecvfd socket, bind to address for IPv4*/
  if ((mcdatarecvfd4 = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("can't open mcdatarecvfd socket!");
    exit(1);
  }
  if (bind(mcdatarecvfd4, (struct sockaddr *) &mcdataaddr4, \
	   sizeof(mcdataaddr4)) < 0) {
    perror("can't bind mcdataaddr v4 to socket!");
    exit(1);
  }
  
  /*get a mcdatarecvfd socket, bind to address for IPv6*/
  if ((mcdatarecvfd = socket(AF_INET6, SOCK_DGRAM, 0)) < 0) {
    perror("can't open mcdatarecvfd socket!");
    exit(1);
  }

  if (bind(mcdatarecvfd, (struct sockaddr *) &mcdataaddr, \
	   sizeof(mcdataaddr)) < 0) {
    perror("can't bind mcdataaddr to socket!");
    exit(1);
  }
  
  /*get a mcrtcprecvfd socket, bind to address for IPv4*/
  if ((mcrtcprecvfd4 = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    perror("can't open mcrtcprecvfd socket!");
    exit(1);
  }

  if (bind(mcrtcprecvfd4, (struct sockaddr *) &mcrtcpaddr4, \
	   sizeof(mcrtcpaddr4)) < 0) {
    perror("can't bind mcrtcpaddr v4 to socket!");
    exit(1);
  }

  /*get a mcrtcprecvfd socket, bind to address for IPv6*/
  if ((mcrtcprecvfd = socket(AF_INET6, SOCK_DGRAM, 0)) < 0) {
    perror("can't open mcrtcprecvfd socket!");
    exit(1);
  }

  if (bind(mcrtcprecvfd, (struct sockaddr *) &mcrtcpaddr, \
	   sizeof(mcrtcpaddr)) < 0) {
    perror("can't bind mcrtcpaddr to socket!");
    exit(1);
  }
  
  /*allow multiple processes per host to read from IPv4 sockets*/
  if (setsockopt(mcdatarecvfd4, SOL_SOCKET, SO_REUSEADDR, &mcdataaddr4, sizeof(mcdataaddr4)) < 0) {
    perror("SO_REUSEADDR setsockopt v4\n");
  }  
  if (setsockopt(mcrtcprecvfd4, SOL_SOCKET, SO_REUSEADDR, &mcrtcpaddr4, sizeof(mcrtcpaddr4)) < 0) {
    perror("SO_REUSEADDR setsockopt v4\n");
  }  
  
  /*allow multiple processes per host to read from IPv6 sockets*/
  if (setsockopt(mcdatarecvfd, SOL_SOCKET, SO_REUSEADDR, &mcdataaddr, sizeof(mcdataaddr)) < 0) {
    perror("SO_REUSEADDR setsockopt\n");
  }  
  if (setsockopt(mcrtcprecvfd, SOL_SOCKET, SO_REUSEADDR, &mcrtcpaddr, sizeof(mcrtcpaddr)) < 0) {
    perror("SO_REUSEADDR setsockopt\n");
  }  
  
  /*set socket options to join multicast group in IPv4*/
  if (setsockopt(mcdatarecvfd4, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mcdatareq4,\
		 sizeof(mcdatareq4)) < 0) {
    perror("can't set socket options to join multicast group v4!");
    exit(1);
  } 
  if (setsockopt(mcrtcprecvfd4, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mcdatareq4,\
		 sizeof(mcdatareq4)) < 0) {
    perror("can't set socket options to join multicast group v4!");
    exit(1);
  } 
  
  /*set socket options to join multicast group*/
  if (setsockopt(mcdatarecvfd, IPPROTO_IPV6, IPV6_JOIN_GROUP, &mcdatareq,\
		 sizeof(mcdatareq)) < 0) {
    perror("can't set socket options to join multicast group!");
    exit(1);
  } 
  if (setsockopt(mcrtcprecvfd, IPPROTO_IPV6, IPV6_JOIN_GROUP, &mcdatareq,\
		 sizeof(mcdatareq)) < 0) {
    perror("can't set socket options to join multicast group!");
    exit(1);
  } 
  
  
  /*now set multicast socket TTL option for IPv4*/
  if (setsockopt(sendfd4, IPPROTO_IP, IP_MULTICAST_TTL, \
		 &ttl4, sizeof(ttl4)) < 0)
    perror("can't set multicast ttl v4 socket option!");
  
  /*now set multicast socket TTL option for IPv6*/
  if (setsockopt(sendfd, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, \
		 &ttl, sizeof(ttl)) < 0)
    perror("can't set multicast ttl socket option!");
  
  /*define the sizes of the address structures*/
  mcdataaddrlen4=sizeof(mcdataaddr4);
  mcrtcpaddrlen4=sizeof(mcrtcpaddr4);
  sourceaddrlen4=sizeof(sourceaddr4);
  
  mcdataaddrlen=sizeof(mcdataaddr);
  mcrtcpaddrlen=sizeof(mcrtcpaddr);
  sourceaddrlen=sizeof(sourceaddr);
    
  
  /*start infinite while loop*/
  /*and check for activity on sockets*/
  
  /*set up the select sytem call parameters*/  
  /*zero out the readfds and writefds lists, then add*/
  /*the file descriptors of interest*/
  while(1) { 
    FD_ZERO(&readfds);
    FD_SET(mcdatarecvfd4, &readfds);
    FD_SET(mcrtcprecvfd4, &readfds);
    FD_SET(mcdatarecvfd, &readfds);
    FD_SET(mcrtcprecvfd, &readfds);
    
    if (debugon >= 1) {
      FD_SET(0, &readfds);
    }
    
    maxfds = mcdatarecvfd +10;
    
    /*check for activity*/
    nfds = select(maxfds, &readfds, NULL, NULL, NULL);
    
    /*if specified on the command line, check for input on stdin*/
    if (debugon >= 1) {
      if (FD_ISSET(0, &readfds)) {
	n2 = read(0,inputbuf, sizeof(inputbuf));
	inputbuf[n2>0? n2-1: 0] = '\0';
 	if (!strcmp(inputbuf,"q")) {
	  programshutdown();
	} 
      }
      fflush(stdout); 
    }
    

    /*1:receive from IPv4, send on IPv6  - data*/ 
    if (FD_ISSET(mcdatarecvfd4, &readfds)) {
      nr = recvfrom(mcdatarecvfd4, mcdatarecvbuf4, MSGBUF_SIZE, 0, (struct sockaddr *) \
		    &sourceaddr4, &sourceaddrlen4);
      if (debugon >= 2){ 
	inet_ntop(AF_INET6,&sourceaddr4.sin_addr.s_addr,addressstring4 ,sizeof(addressstring4));
	printf("\nreading from mcdatarecvfd4, got data from %s\n", addressstring4);
      }
      
      if (sourceaddr4.sin_addr.s_addr==localsendaddr4.sin_addr.s_addr){
	chksrc = 0;
	if (debugon >= 2) 
	  printf("don't retransmit multicastv4 sourced from gateway machine\n");
      } else {
	chksrc = 1;
	if (debugon >= 2) 
	  printf("retransmit to multicastv6 address\n");
      }
      
      if (chksrc) {
	mcdatarecvfromcalls4 = mcdatarecvfromcalls4++;
	mcdatatotalbytesrecv4 = mcdatatotalbytesrecv4 + nr;
	if (nr < 0)
	  printf ("mcdatarecvfd4:recvfrom over multicast v4 address error!(1)\n");     
	
	/*sent to the multicast group*/
	if (debugon >= 2){
	  inet_ntop(AF_INET6,&mcdataaddr.sin6_addr,addressstring ,sizeof(addressstring));
	  printf("sending to %s\n", addressstring);
	}
	ns = sendto(sendfd, mcdatarecvbuf4, nr, 0, (struct sockaddr *)&mcdataaddr, \
		    sizeof(mcdataaddr));
	mcdatasendtocalls = mcdatasendtocalls++;
	mcdatatotalbytessent = mcdatatotalbytessent + ns;
	if (ns < 0)
	  printf("sendfd:sendto over multicast v6 address error!(2)\n");
      }
    }
    
    /*2:receive from IPv4, send on IPv6  - rtcp*/
    if (FD_ISSET(mcrtcprecvfd4, &readfds)) {
      nr = recvfrom(mcrtcprecvfd4, mcrtcprecvbuf4, MSGBUF_SIZE, 0, (struct sockaddr *) \
		    &sourceaddr4, &sourceaddrlen4);
      if (debugon >= 2) {
	inet_ntop(AF_INET,&sourceaddr4.sin_addr.s_addr,addressstring4 ,sizeof(addressstring4));
	printf("\nreading from mcrtcprecvfd4, got data from %s\n", addressstring4);
      }
      
      if (sourceaddr4.sin_addr.s_addr==localsendaddr4.sin_addr.s_addr){
	chksrc = 0;
	if (debugon >= 2) 
	  printf("don't retransmit multicastv4 sourced from gateway machine\n");
      } else {
	chksrc = 1;
	if (debugon >= 2) 
	  printf("retransmit to unicast address\n");
      }
         
      if (chksrc) {
	mcrtcprecvfromcalls4 = mcrtcprecvfromcalls4++;
	mcrtcptotalbytesrecv4 = mcrtcptotalbytesrecv4 + nr; 
	if (nr < 0)
	  printf ("crtcprecvfd4 :recvfrom over multicast v4 address error!(3)\n");
	
	/*sent to the multicast group*/
	if (debugon >= 2) {
	  inet_ntop(AF_INET6,&mcrtcpaddr.sin6_addr,addressstring ,sizeof(addressstring));
	  printf("sending to %s\n", addressstring);
	}
	ns = sendto(sendfd, mcrtcprecvbuf4, nr, 0, (struct sockaddr *)&mcrtcpaddr, \
		    sizeof(mcrtcpaddr));
	mcrtcpsendtocalls = mcrtcpsendtocalls++;
	mcrtcptotalbytessent = mcrtcptotalbytessent + ns;
	if (ns < 0)
	  printf("sendfd:sendto over multicast v6 address error!(4)\n");
      }
    }
    
    /*3:receive from IPv6, send on IPv4 - data*/ 
    if (FD_ISSET(mcdatarecvfd, &readfds)) {
      nr = recvfrom(mcdatarecvfd, mcdatarecvbuf, MSGBUF_SIZE, 0, (struct sockaddr *) \
		    &sourceaddr, &sourceaddrlen);
      if (debugon >= 2)  {
	inet_ntop(AF_INET6,&sourceaddr.sin6_addr,addressstring ,sizeof(addressstring));
	printf("\nreading from mcdatarecvfd, got data from=%s\n",addressstring);
	inet_ntop(AF_INET,&localsendaddr4.sin_addr.s_addr,addressstring ,sizeof(addressstring));
	printf("localsendaddr=%s\n",addressstring);
      }
      if (IN6_ARE_ADDR_EQUAL(&sourceaddr.sin6_addr, &localsendaddr.sin6_addr)){
	chksrc = 0;
	if (debugon >= 2) 
	  printf("don't retransmit multicast IPv6 sourced from gateway machine\n");
      } else {
	chksrc = 1;
	if (debugon >= 2) 
	  printf("retransmit to IPv4 address\n");
      }
      if (chksrc != 0 ) {
	mcdatarecvfromcalls = mcdatarecvfromcalls++;
	mcdatatotalbytesrecv = mcdatatotalbytesrecv + nr;
	if (nr < 0)
	  printf ("mcdatarecvfd:recvfrom over multicast v6 address error!(5)\n");
	
	/*now send to IPv4*/
	if (multicastaddress4.s_addr!=INADDR_ANY) {
	  if (debugon >= 2) {
	    inet_ntop(AF_INET,&mcdataaddr4.sin_addr.s_addr,addressstring4 ,sizeof(addressstring4));
	    printf("sending to %s\n", addressstring4);
	  } 
	  
	  ns = sendto(sendfd4, mcdatarecvbuf, nr, 0, (struct sockaddr *)&mcdataaddr4, \
		      sizeof(mcdataaddr4));
	  
	  mcdatasendtocalls4 = mcdatasendtocalls4++;
	  mcdatatotalbytessent4 = mcrtcptotalbytessent4++;
	} else {
	  if (debugon >= 2) printf("not resending to ORIGINATOR! or array entry = 0\n");
	}
      }//end of if (chksrc == 0)
    }
    
    /*4:receive from IPv6, send on IPv4 - rtcp*/ 
    if (FD_ISSET(mcrtcprecvfd, &readfds)) {
      nr = recvfrom(mcrtcprecvfd, mcrtcprecvbuf, MSGBUF_SIZE, 0, (struct sockaddr *) \
		    &sourceaddr, &sourceaddrlen);
      if (debugon >= 2)  {
	inet_ntop(AF_INET6,&sourceaddr.sin6_addr,addressstring ,sizeof(addressstring));
	printf("\nreading from mcrtcprecvfd, got data from=%s\n",addressstring);
	inet_ntop(AF_INET,&localsendaddr4.sin_addr.s_addr,addressstring ,sizeof(addressstring));
	printf("localsendaddr=%s\n",addressstring);
      }
      if (IN6_ARE_ADDR_EQUAL(&sourceaddr.sin6_addr, &localsendaddr.sin6_addr)){
	chksrc = 0;
	if (debugon >= 2)
	  printf("don't retransmit\n");
      } else {
	chksrc = 1;
	if (debugon >= 2)
	  printf("retransmit to multicast v4 address\n");
      }
      if (chksrc != 0 ) {
	mcrtcprecvfromcalls = mcrtcprecvfromcalls++;
	mcrtcptotalbytesrecv = mcrtcptotalbytesrecv + nr;
	if (nr < 0)
	  printf ("mcrtcprecvfd:recvfrom over multicast address error!(6)\n");
	
	/*now  send to IPv4 member*/
	if (multicastaddress4.s_addr!=INADDR_ANY) {
	  if (debugon >= 2){
	    inet_ntop(AF_INET,&mcrtcpaddr4.sin_addr.s_addr,addressstring4 ,sizeof(addressstring4));
	    printf("sending to %s\n", addressstring4);
	  }
	  ns = sendto(sendfd4, mcrtcprecvbuf, nr, 0, (struct sockaddr *)&mcrtcpaddr4, \
		      sizeof(mcrtcpaddr4));
	  mcrtcpsendtocalls4 = mcrtcpsendtocalls4++;
	  mcrtcptotalbytessent4 = mcrtcptotalbytessent4++;
	} else {
	  if (debugon >= 4) printf("not resending to ORIGINATOR! or array entry = 0\n");
	}
      }//end of if (chksrc == 0)
    }
    
  } //end of while loop
} //end of main


int programshutdown () {
  extern int mcdatatotalbytesrecv4;
  extern int mcdatatotalbytesrecv;
  extern int mcdatarecvfromcalls4;
  extern int mcdatarecvfromcalls;
  extern int mcdatatotalbytessent4;
  extern int mcdatatotalbytessent;
  extern int mcdatasendtocalls4;
  extern int mcdatasendtocalls;
  extern int mcrtcptotalbytesrecv4;
  extern int mcrtcptotalbytesrecv;
  extern int mcrtcprecvfromcalls4;
  extern int mcrtcprecvfromcalls;
  extern int mcrtcptotalbytessent4;
  extern int mcrtcptotalbytessent;
  extern int mcrtcpsendtocalls4;
  extern int mcrtcpsendtocalls;
  
  /*print out debug information*/
printf("number of mc4 data bytes received = %d\n", mcdatatotalbytesrecv4);
printf("number of mc4 data recvfromcalls = %d\n", mcdatarecvfromcalls4);
printf("number of mc data bytes sent = %d\n", mcdatatotalbytessent);
printf("number of mc data sendtocalls = %d\n", mcdatasendtocalls);
printf("--------------------------------\n");
printf("number of mc4 rtcp bytes received = %d\n", mcrtcptotalbytesrecv4);
printf("number of mc4 rtcp recvfromcalls = %d\n", mcrtcprecvfromcalls4);
printf("number of mc rtcp bytes sent = %d\n", mcrtcptotalbytessent);
printf("number of mc rtcp sendtocalls = %d\n", mcrtcpsendtocalls);
printf("--------------------------------\n");
printf("number of mc data bytes received = %d\n", mcdatatotalbytesrecv);
printf("number of mc data recvfromcalls = %d\n", mcdatarecvfromcalls);
printf("number of mc4 data bytes sent = %d\n", mcdatatotalbytessent4);
printf("number of mc4 data sendtocalls = %d\n", mcdatasendtocalls4);
printf("--------------------------------\n");
printf("number of mc rtcp bytes received = %d\n", mcrtcptotalbytesrecv);
printf("number of mc rtcp recvfromcalls = %d\n", mcrtcprecvfromcalls);
printf("number of mc4 rtcp bytes sent = %d\n", mcrtcptotalbytessent4);
printf("number of mc4 rtcp sendtocalls = %d\n", mcrtcpsendtocalls4);
exit(0);
}
 


